#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.
""" SKALE Allocator Core Escrow methods """

from enum import IntEnum

from skale.contracts.base_contract import BaseContract, transaction_method
from skale.transactions.result import TxRes
from skale.utils.helper import format_fields


PLAN_FIELDS = [
    'totalVestingDuration',
    'vestingCliff',
    'vestingIntervalTimeUnit',
    'vestingInterval',
    'isDelegationAllowed',
    'isTerminatable'
]

BENEFICIARY_FIELDS = [
    'status',
    'planId',
    'startMonth',
    'fullAmount',
    'amountAfterLockup'
]

MAX_NUM_OF_PLANS = 9999
MAX_NUM_OF_BENEFICIARIES = 9999


class TimeUnit(IntEnum):
    DAY = 0
    MONTH = 1
    YEAR = 2


class BeneficiaryStatus(IntEnum):
    UNKNOWN = 0
    CONFIRMED = 1
    ACTIVE = 2
    TERMINATED = 3


class Allocator(BaseContract):
    def is_beneficiary_registered(self, beneficiary_address: str) -> bool:
        """Confirms whether the beneficiary is registered in a Plan.

        :returns: Boolean value
        :rtype: bool
        """
        return self.contract.functions.isBeneficiaryRegistered(beneficiary_address).call()

    def is_delegation_allowed(self, beneficiary_address: str) -> bool:
        return self.contract.functions.isDelegationAllowed(beneficiary_address).call()

    def is_vesting_active(self, beneficiary_address: str) -> bool:
        return self.contract.functions.isVestingActive(beneficiary_address).call()

    def get_escrow_address(self, beneficiary_address: str) -> str:
        return self.contract.functions.getEscrowAddress(beneficiary_address).call()

    @transaction_method
    def add_plan(
            self,
            vesting_cliff: int,
            total_vesting_duration: int,
            vesting_interval_time_unit: TimeUnit,
            vesting_interval: int,
            can_delegate: bool,
            is_terminatable: bool
    ) -> TxRes:
        return self.contract.functions.addPlan(
            vestingCliff=vesting_cliff,
            totalVestingDuration=total_vesting_duration,
            vestingIntervalTimeUnit=vesting_interval_time_unit.value,
            vestingInterval=vesting_interval,
            canDelegate=can_delegate,
            isTerminatable=is_terminatable
        )

    @transaction_method
    def connect_beneficiary_to_plan(
            self,
            beneficiary_address: str,
            plan_id: int,
            start_month: int,
            full_amount: int,
            lockup_amount: int,
    ) -> TxRes:
        return self.contract.functions.connectBeneficiaryToPlan(
            beneficiary=beneficiary_address,
            planId=plan_id,
            startMonth=start_month,
            fullAmount=full_amount,
            lockupAmount=lockup_amount
        )

    @transaction_method
    def start_vesting(self, beneficiary_address: str) -> TxRes:
        return self.contract.functions.startVesting(beneficiary_address)

    @transaction_method
    def stop_vesting(self, beneficiary_address: str) -> TxRes:
        return self.contract.functions.stopVesting(beneficiary_address)

    @transaction_method
    def grant_role(self, role: bytes, address: str) -> TxRes:
        return self.contract.functions.grantRole(role, address)

    def vesting_manager_role(self) -> bytes:
        return self.contract.functions.VESTING_MANAGER_ROLE().call()

    def has_role(self, role: bytes, address: str) -> bool:
        return self.contract.functions.hasRole(role, address).call()

    def __get_beneficiary_plan_params_raw(self, beneficiary_address: str):
        return self.contract.functions.getBeneficiaryPlanParams(beneficiary_address).call()

    @format_fields(BENEFICIARY_FIELDS)
    def get_beneficiary_plan_params_dict(self, beneficiary_address: str) -> dict:
        return self.__get_beneficiary_plan_params_raw(beneficiary_address)

    def get_beneficiary_plan_params(self, beneficiary_address: str) -> dict:
        plan_params = self.get_beneficiary_plan_params_dict(beneficiary_address)
        plan_params['statusName'] = BeneficiaryStatus(plan_params['status']).name
        return plan_params

    def __get_plan_raw(self, plan_id: int):
        return self.contract.functions.getPlan(plan_id).call()

    @format_fields(PLAN_FIELDS)
    def get_plan(self, plan_id: int) -> dict:
        return self.__get_plan_raw(plan_id)

    def get_all_plans(self) -> dict:
        plans = []
        for i in range(1, MAX_NUM_OF_PLANS):
            try:
                plan = self.get_plan(i)
                plan['planId'] = i
                plans.append(plan)
            except ValueError:
                break
        return plans



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

import functools
import json
import logging
import time
import urllib

import requests
from hexbytes import HexBytes
from eth_account.datastructures import AttributeDict

from skale.wallets.common import BaseWallet
from skale.utils.exceptions import RPCWalletError


logger = logging.getLogger(__name__)

ROUTES = {
    'sign': '/sign',
    'sign_and_send': '/sign-and-send',
    'sign_hash': '/sign-hash',
    'address': '/address',
    'public_key': '/public-key',
}

ATTEMPTS = 10
TIMEOUTS = [2 ** p for p in range(ATTEMPTS)]
SGX_UNREACHABLE_MESSAGE = 'Sgx server is unreachable'


def rpc_request(func):
    @functools.wraps(func)
    def wrapper(self, route, *args, **kwargs):
        data, error = None, None
        for i, timeout in enumerate(TIMEOUTS):
            logger.info(f'Sending request to tm for {route}. Try {i}')
            try:
                response = func(self, route, *args, **kwargs).json()
                data, error = response.get('data'), response.get('error')
            except Exception as err:
                error = 'RPC request failed'
                logger.error(error, exc_info=err)

            if not error or not self._retry_if_failed:
                break

            logger.info(f'Sleeping {timeout}s ...')
            time.sleep(timeout)

        if error is not None:
            raise RPCWalletError(error)
        return data
    return wrapper


class RPCWallet(BaseWallet):
    def __init__(self, url, retry_if_failed=False):
        self._url = url
        self._retry_if_failed = retry_if_failed

    def _construct_url(self, host, url):
        return urllib.parse.urljoin(host, url)

    @rpc_request
    def _post(self, route, data):
        request_url = self._construct_url(self._url, route)
        return requests.post(request_url, json=data)

    @rpc_request
    def _get(self, route, data=None):
        request_url = self._construct_url(self._url, route)
        return requests.get(request_url, data=data)

    def _compose_tx_data(self, tx_dict):
        return {
            'transaction_dict': json.dumps(tx_dict)
        }

    def sign(self, tx_dict):
        data = self._post(ROUTES['sign'], self._compose_tx_data(tx_dict))
        return AttributeDict(data)

    def sign_and_send(self, tx_dict):
        data = self._post(ROUTES['sign_and_send'], self._compose_tx_data(tx_dict))
        return data['transaction_hash']

    def sign_hash(self, unsigned_hash: str):
        data = self._post(ROUTES['sign_hash'], {'unsigned_hash': unsigned_hash})
        return AttributeDict({
            'messageHash': HexBytes(data['messageHash']),
            'r': data['r'],
            's': data['s'],
            'v': data['v'],
            'signature': HexBytes(data['signature']),
        })

    @property
    def address(self):
        data = self._get(ROUTES['address'])
        return data['address']

    @property
    def public_key(self):
        data = self._get(ROUTES['public_key'])
        return data['public_key']



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.
""" SKALE base contract class """

import logging
from functools import wraps

from web3 import Web3

from skale.transactions.result import (is_success,
                                       is_success_or_not_performed,
                                       TxRes, check_balance_and_gas)
from skale.transactions.tools import post_transaction, make_dry_run_call
from skale.utils.web3_utils import wait_for_receipt_by_blocks, wait_for_confirmation_blocks
from skale.utils.account_tools import account_eth_balance_wei


logger = logging.getLogger(__name__)


def transaction_method(transaction):
    @wraps(transaction)
    def wrapper(self, *args, wait_for=True,
                wait_timeout=4, blocks_to_wait=50, gas_limit=None,
                gas_price=None, nonce=None, dry_run_only=False, skip_dry_run=False,
                raise_for_status=True, confirmation_blocks=0, **kwargs):
        method = transaction(self, *args, **kwargs)
        dry_run_result, tx_hash, receipt = None, None, None

        # Make dry_run
        if not skip_dry_run:
            dry_run_result = make_dry_run_call(self.skale, method, gas_limit)
            success_dry_run = is_success(dry_run_result)
            if success_dry_run:
                gas_limit = dry_run_result['payload']

        # Check balance
        balance = account_eth_balance_wei(self.skale.web3,
                                          self.skale.wallet.address)
        gas_price = gas_price or self.skale.gas_price
        balance_check_result = check_balance_and_gas(balance, gas_price, gas_limit)
        rich_enough = is_success(balance_check_result)

        # Send transaction
        should_send_transaction = not dry_run_only and \
            is_success_or_not_performed(dry_run_result)

        if rich_enough and should_send_transaction:
            tx_hash = post_transaction(
                self.skale.wallet, method, gas_limit,
                gas_price, nonce
            )
            if wait_for:
                receipt = wait_for_receipt_by_blocks(
                    self.skale.web3,
                    tx_hash,
                    timeout=wait_timeout,
                    blocks_to_wait=blocks_to_wait
                )
            if confirmation_blocks:
                wait_for_confirmation_blocks(self.skale.web3, confirmation_blocks)

        tx_res = TxRes(dry_run_result, balance_check_result, tx_hash, receipt)

        if raise_for_status:
            tx_res.raise_for_status()
        return tx_res

    return wrapper


class BaseContract:
    def __init__(self, skale, name, address, abi):
        self.skale = skale
        self.name = name
        self.address = Web3.toChecksumAddress(address)
        self.contract = skale.web3.eth.contract(address=self.address, abi=abi)



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

import logging
import time
from functools import partial, wraps

from skale.transactions.result import (
    DryRunFailedError,
    InsufficientBalanceError,
    TransactionFailedError, TxRes
)
from skale.utils.constants import GAS_LIMIT_COEFFICIENT
from skale.utils.exceptions import RPCWalletError
from skale.utils.web3_utils import get_eth_nonce

from web3._utils.transactions import get_block_gas_limit

logger = logging.getLogger(__name__)


def make_dry_run_call(skale, method, gas_limit=None) -> dict:
    opts = {
        'from': skale.wallet.address,
    }
    logger.info(
        f'Dry run tx: {method.fn_name}, '
        f'sender: {skale.wallet.address}, '
        f'wallet: {skale.wallet.__class__.__name__}, '
    )
    try:
        if gas_limit:
            estimated_gas = gas_limit
            opts.update({'gas': gas_limit})
            method.call(opts)
        else:
            estimated_gas = estimate_gas(skale.web3, method, opts)
        logger.info(f'Estimated gas for {method.fn_name}: {estimated_gas}')
    except Exception as err:
        logger.error('Dry run for method failed with error', exc_info=err)
        return {'status': 0, 'error': str(err)}

    return {'status': 1, 'payload': estimated_gas}


def estimate_gas(web3, method, opts):
    block_gas_limit = get_block_gas_limit(web3)
    estimated_gas = method.estimateGas(opts)
    normalized_estimated_gas = int(estimated_gas * GAS_LIMIT_COEFFICIENT)
    if normalized_estimated_gas > block_gas_limit:
        logger.warning(f'Estimate gas for {method.fn_name} exceeds block gas limit')
        return block_gas_limit
    return normalized_estimated_gas


def build_tx_dict(method, gas_limit, gas_price=None, nonce=None):
    tx_dict_fields = {
        'gas': gas_limit,
        'nonce': nonce
    }
    if gas_price is not None:
        tx_dict_fields.update({'gasPrice': gas_price})

    return method.buildTransaction(tx_dict_fields)


def sign_and_send(web3, method, gas_amount, wallet) -> hash:
    nonce = get_eth_nonce(web3, wallet.address)
    tx_dict = build_tx_dict(method, gas_amount, nonce)
    signed_tx = wallet.sign(tx_dict)
    return web3.eth.sendRawTransaction(signed_tx.rawTransaction)


def post_transaction(wallet, method, gas_limit, gas_price=None, nonce=None) -> str:
    logger.info(
        f'Tx: {method.fn_name}, '
        f'sender: {wallet.address}, '
        f'wallet: {wallet.__class__.__name__}, '
        f'gasLimit: {gas_limit}, '
        f'gasPrice: {gas_price}'
    )
    tx_dict = build_tx_dict(method, gas_limit, gas_price, nonce)
    tx_hash = wallet.sign_and_send(tx_dict)
    return tx_hash


def send_eth(web3, account, amount, wallet):
    eth_nonce = get_eth_nonce(web3, wallet.address)
    logger.info(f'Transaction nonce {eth_nonce}')
    txn = {
        'to': account,
        'value': amount,
        'gasPrice': web3.eth.gasPrice,
        'gas': 22000,
        'nonce': eth_nonce
    }
    signed_txn = wallet.sign(txn)
    tx = web3.eth.sendRawTransaction(signed_txn.rawTransaction)

    logger.info(
        f'ETH transfer {wallet.address} => {account}, {amount} wei,'
        f'tx: {web3.toHex(tx)}'
    )
    return tx


def retry_tx(tx=None, *, max_retries=3, timeout=-1):
    if tx is None:
        return partial(retry_tx, max_retries=3, timeout=timeout)

    @wraps(tx)
    def wrapper(*args, **kwargs):
        return run_tx_with_retry(
            tx, *args,
            max_retries=max_retries,
            retry_timeout=timeout, **kwargs
        )
    return wrapper


def run_tx_with_retry(transaction, *args, max_retries=3,
                      retry_timeout=-1,
                      **kwargs) -> TxRes:
    success = False
    attempt = 0
    tx_res = None
    exp_timeout = 1
    while not success and attempt < max_retries:
        try:
            tx_res = transaction(*args, **kwargs)
            tx_res.raise_for_status()
        except (TransactionFailedError, DryRunFailedError, RPCWalletError) as err:
            logger.error(f'Tx attempt {attempt}/{max_retries} failed',
                         exc_info=err)
            timeout = exp_timeout if retry_timeout < 0 else exp_timeout
            time.sleep(timeout)
            exp_timeout *= 2
        except InsufficientBalanceError as err:
            logger.error('Sender balance is too low', exc_info=err)
            raise err
        else:
            success = True
        attempt += 1
    if success:
        logger.info(f'Tx {transaction.__name__} completed '
                    f'after {attempt}/{max_retries} retries')
    else:
        logger.error(
            f'Tx {transaction.__name__} failed after '
            f'{max_retries} retries')
    return tx_res



# flake8: noqa: E402

import sys

if sys.version_info < (3, 7):
    raise EnvironmentError("Python 3.7 or above is required")

from skale.skale_manager import SkaleManager
from skale.skale_manager import SkaleManager as Skale  # todo: deprecated naming, will be removed in skale.py v5

from skale.skale_allocator import SkaleAllocator



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from skale.utils.exceptions import IncompatibleAbiError


def get_abi_key_name(contract_name):
    # todo: tmp fix for inconsistent contract names
    if contract_name == 'manager' or contract_name == 'token':
        return f'skale_{contract_name}_abi'
    if contract_name == 'dkg':
        return 'skale_d_k_g_abi'
    return f'{contract_name}_abi'


def get_address_key_name(contract_name):
    if contract_name == 'manager' or contract_name == 'token':
        return f'skale_{contract_name}_address'
    if contract_name == 'dkg':
        return 'skale_d_k_g_address'
    return f'{contract_name}_address'


def get_abi_key(abi, key):
    try:
        return abi[key]
    except KeyError:
        raise IncompatibleAbiError(key)


def get_contract_address_by_name(abi, name):
    return get_abi_key(abi, get_address_key_name(name))


def get_contract_abi_by_name(abi, name):
    return get_abi_key(abi, get_abi_key_name(name))



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

import abc
import logging

from web3 import Web3
from web3.middleware import geth_poa_middleware

from skale.wallets import BaseWallet
from skale.utils.abi_utils import get_contract_address_by_name, get_contract_abi_by_name
from skale.utils.constants import GAS_PRICE_COEFFICIENT
from skale.utils.exceptions import InvalidWalletError, EmptyWalletError
from skale.utils.web3_utils import get_provider

logger = logging.getLogger(__name__)


class EmptyPrivateKey(Exception):
    pass


class SkaleBase:
    __metaclass__ = abc.ABCMeta

    def __init__(self, endpoint, abi_filepath, wallet=None, provider_timeout=30):
        logger.info(f'Initing skale.py, endpoint: {endpoint}, wallet: {type(wallet).__name__}')
        provider = get_provider(endpoint, timeout=provider_timeout)
        self._abi_filepath = abi_filepath
        self._endpoint = endpoint
        self.web3 = Web3(provider)
        self.web3.middleware_onion.inject(
            geth_poa_middleware, layer=0)  # todo: may cause issues
        self.__contracts = {}
        if wallet:
            self.wallet = wallet
        self.init_contracts()

    @property
    def gas_price(self):
        return self.web3.eth.gasPrice * GAS_PRICE_COEFFICIENT

    @property
    def wallet(self):
        if not self._wallet:
            raise EmptyWalletError('No wallet provided')
        return self._wallet

    @wallet.setter
    def wallet(self, wallet):
        if issubclass(type(wallet), BaseWallet):
            self._wallet = wallet
        else:
            raise InvalidWalletError(f'Wrong wallet class: {type(wallet).__name__}. \
                                       Must be one of the BaseWallet subclasses')

    def __is_debug_contracts(self, abi):
        return abi.get('time_helpers_with_debug_address', None)

    @abc.abstractmethod
    def init_contracts(self):
        return

    def __init_contracts_from_info(self, abi, contracts_info):
        for name in contracts_info:
            info = contracts_info[name]
            if info.upgradeable:
                self.init_upgradeable_contract(info, abi)
            else:
                self.add_lib_contract(info.name, info.contract_class, abi)

    def init_upgradeable_contract(self, contract_info, abi):
        address = self.get_contract_address(contract_info.contract_name)
        self.add_lib_contract(contract_info.name, contract_info.contract_class,
                              abi, address)

    def add_lib_contract(self, name, contract_class, abi, contract_address=None):
        address = contract_address or get_contract_address_by_name(
            abi, name)
        logger.debug(f'Fetching abi for {name}, address {address}')
        contract_abi = get_contract_abi_by_name(abi, name)
        self.add_contract(name, contract_class(
            self, name, address, contract_abi))

    def add_contract(self, name, skale_contract):
        logger.debug(f'Init contract: {name}')
        self.__contracts[name] = skale_contract

    def get_contract_address(self, name):
        return self.contract_manager.get_contract_address(name)

    def get_contract_by_name(self, name):
        return self.__contracts[name]

    def __getattr__(self, name):
        if name not in self.__contracts:
            return None
        return self.get_contract_by_name(name)
